/* 
 * au_audio.c
 * 
 * Aureal Vortex AU88xx audio driver module
 * Wave Audio
 * 
 * Copyright (c) 1999 Aureal Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by   
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *   
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Some of this code is based on the es1371 driver, which is
 * Copyright (C) 1998  Thomas Sailer (sailer@ife.ee.ethz.ch)
 *
 * Revision History:
 *
 * 28-07-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Added select routines, plays nicely with SDL and OpenAL now
 *
 * 16-08-2000 Warren Chartier <icebalm@bigfoot.com>
 * 	Fixed DSP_SETTRIGER, bad bug which would cause only 1 frag to be used
 *
 * 19-08-2000 Nikos Kalogridis
 *	Split vortex.c into multiple files for easier maintenance
 *
 * 22-08-2000 Steve Butler
 *	Fixed kernel oops from wav file playback
 *
 * 23-08-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Reformatted code
 *
 * 25-08-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Reversed some of my changes, fixed Quake3
 *
 */

#define __NO_VERSION__
#include "au_vortex.h"
#include <linux/version.h>

static ssize_t au_audio_read(struct file *file, char *buffer, size_t count, loff_t *ppos);
static ssize_t au_audio_write(struct file *file, const char *buffer, size_t count, loff_t *ppos);
static unsigned int au_audio_poll(struct file *file, struct poll_table_struct *wait);
static int au_audio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int au_audio_mmap(struct file *file, struct vm_area_struct *vma);
static int au_audio_open(struct inode *inode, struct file *file);
static int au_audio_release(struct inode *inode, struct file *file);
extern void write_mixer(au_device_t *dev, unsigned int chan, int val);

/* Audio file operations */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
static struct file_operations audio_fops =
{
	owner:		THIS_MODULE,
	llseek:		au_llseek,
	read:		au_audio_read,
	write:		au_audio_write,
	poll:		au_audio_poll,
	ioctl:		au_audio_ioctl,
	mmap:		au_audio_mmap,
	open:		au_audio_open,
	release:	au_audio_release,
};
#else
static struct file_operations audio_fops = {
        au_llseek,
        au_audio_read,
        au_audio_write,
        NULL,                   /* readdir */   
        au_audio_poll,
        au_audio_ioctl,
        au_audio_mmap,
        au_audio_open,
        NULL,                   /* flush */
        au_audio_release,
        NULL,                   /* fsync */
        NULL,                   /* fasync */
        NULL,                   /* check media change */
        NULL,                   /* revalidate */
        NULL,                   /* lock */
};
#endif

static int add_wave_device(au_device_t *dev, struct file *file, int dev_type);
static int get_frags(wave_stream_t *stream, unsigned long count);
static void free_wave_device(au_device_t *dev, wave_device_t *wdev);
static wave_device_t* find_wave_device(au_device_t *dev, struct file *file);
static void remove_wave_device(au_device_t *dev, wave_device_t *wdev);
static unsigned long receive_mem(au_device_t *dev, wave_stream_t *stream, char *mem, unsigned long max_size);
static unsigned long send_mem(au_device_t *dev, wave_stream_t *stream, const char *mem, unsigned long size);
static void set_data_format(wave_stream_t *stream, int val);
static void set_asp_format(wave_stream_t *stream);
static void stop_wave(au_device_t *dev, wave_stream_t *stream);
static void drain_playback(au_device_t *dev, wave_device_t *wdev, int nonblock);


/********************************/
/* Wave audio support		*/
/********************************/

int au_dsp_register(au_device_t *dev)
{
	int status;
	
	/* register dsp device with soundcore */
	PDEBUG("au_dsp_register\n");
	status = dev->dev_audio = register_sound_dsp(&audio_fops, -1);
	return (status < 0) ? status : 0;
}

int au_dsp_unregister(au_device_t *dev) 
{
	PDEBUG("au_dsp_unregister\n");
	unregister_sound_dsp(dev->dev_audio);
	return 0;
}	

int au_audio_register(au_device_t *dev)
{
	int status;
	
	/* register audio device with soundcore */
	PDEBUG("au_audio_register\n");
	status = dev->dev_sun = register_sound_special(&audio_fops, 4);
	return (status < 0) ? status : 0;
}

int au_audio_unregister(au_device_t *dev) {
	PDEBUG("au_audio_unregister\n");
	unregister_sound_special(dev->dev_sun);
	return 0;
}	

static ssize_t au_audio_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	au_device_t *dev = file->private_data;
	wave_device_t *wdev;
	wave_stream_t *stream;
	size_t received = 0;
	unsigned long flags = 0;
	size_t xfer;

	PSHOW("au_audio_read, count=%d\n", count);
	wdev = find_wave_device(dev, file);
	if (!wdev) {
		PWARN("could not find wave device for file = 0x%p\n", file);
		return -ENXIO;
	}
	stream = &wdev->record_stream;
	/* assert(stream->wave) */
	if(ppos != &file->f_pos)
		return -ESPIPE;
	if(stream->mapped)
		return -ENXIO;
	if(!access_ok(VERIFY_WRITE, buffer, count))
		return -EFAULT;
	if(!stream->active) {
		set_asp_format(stream);
		write_mixer(dev, SOUND_MIXER_RECLEV, dev->mix_vol[SOUND_MIXER_RECLEV]);
		PSHOW("Record stream write_mixer called: SOUND_MIXER_RECLEV = %ix\n", dev->mix_vol[SOUND_MIXER_RECLEV]);
	}	

	while (received < count) {
		/*
		 * If inactive, receive_mem will add the first buffers.
		 * Otherwise, it will copy completed buffers to user memory
		 * and then queue up another one.
		 */
		xfer = receive_mem(dev, stream, buffer + received, count - received);
		received += xfer;
		if (!stream->active) {
			spin_lock_irqsave(&dev->irq_lock, flags);
			StartWave(stream->wave);
			stream->active = TRUE;
			spin_unlock_irqrestore(&dev->irq_lock, flags);
		}

		if (received < count) {
			if (file->f_flags & O_NONBLOCK)
				return received ? (ssize_t)received : -EAGAIN;
			interruptible_sleep_on(&stream->io_wait);
			if (signal_pending(current))
				return received ? (ssize_t)received : -ERESTARTSYS;
		}
	}
	return received ? (ssize_t)received : -EFAULT;
}

static ssize_t au_audio_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
	au_device_t *dev = file->private_data;
	wave_device_t *wdev;
	wave_stream_t *stream;
	unsigned long flags = 0;
	size_t sent = 0;

	PSHOW("au_audio_write() called\n");
	wdev = find_wave_device(dev, file);
	if (!wdev) {
		PWARN("Could not find wave device for file = 0x%p\n", file);
		return -ENXIO;
	}
	stream = &wdev->play_stream;
	/* assert(stream->wave) */

	if(ppos != &file->f_pos)
		return -ESPIPE;
	if(stream->mapped)
		return -ENXIO;

	if(!access_ok(VERIFY_READ, buffer, count))
		return -EFAULT;
	/* asp core format must be set before first add buffer */
	if(!stream->active)
		set_asp_format(stream);			
	PSHOW("au_audio_write() : entry point.\n");
	if(!count)
		return(0);    //[BUGFIX]
	while(sent < count) {
		/*
		 * Copy as much as we can into the dma buffers.
		 */
		size_t xfer = send_mem(dev, stream, buffer + sent, count - sent);
		PSHOW("au_audio_write() : loop.\n");
		sent += xfer;
		stream->total_sent += xfer;
		PSHOW("xfer  = %i\n",  xfer);
		PSHOW("sent  = %i\n",  sent);
		PSHOW("count = %i\n",  count);
		PSHOW("playd = %li\n", stream->total_played);
		PSHOW("total = %li\n", stream->total_sent);
		/* If pre-rolling, we can't allow blocking since app must send trigger. */
		if(stream->pre_roll) {
			if(sent < count) {
				PSHOW("au_audio_write() : exit was 1\n");
				return sent ? (ssize_t)sent : -EAGAIN;
			} else {
				PSHOW("au_audio_write() : exit was 2\n");
				break;
			}
		}
		if(stream->active) {
			/* DMA is running, so we must block if not enough room for user data. */
			if(sent < count) {	
				if(file->f_flags & O_NONBLOCK) {
					PSHOW("au_audio_write() : exit was 3\n");
					return sent ? (ssize_t)sent : -EAGAIN;
				}
				interruptible_sleep_on(&stream->io_wait);
				if(signal_pending(current)) {
					/* app was killed */
					PSHOW("au_audio_write() : exit was 4 (app killed)\n");
					return sent ? (ssize_t)sent : -ERESTARTSYS;
				}
			}
		} else {
			/*
			 * DMA is not running, so start it if we have enough data. If not,
			 * app will have to send DSP_SYNC (or a close) to start.
			 */
			if (stream->total_sent >= stream->frag_size) {
				spin_lock_irqsave(&dev->irq_lock, flags);   
				StartWave(stream->wave);
				stream->active = TRUE;
				spin_unlock_irqrestore(&dev->irq_lock, flags);
			} else {
				if(sent < count) {
					/*
					 * Inactive, but couldn't copy even a frag size.
					 * Shouldn't happen.
					 */
					PSHOW("au_audio_write() : exit was 5\n");
					return sent ? (ssize_t)sent : -EAGAIN;
				}
			}
		} /* if (stream->active) */
	}
	PSHOW("au_audio_write() : exit was 6 (normal)\n");
	return sent ? (ssize_t)sent : -EFAULT;
}

static unsigned int au_audio_poll(struct file *file, struct poll_table_struct *wait)
{
	au_device_t *dev = file->private_data;
	wave_device_t *wdev;
	unsigned long b_count = 0;
	unsigned long flags;
	unsigned int mask = 0;
	int i;

	PSHOW("au_audio_poll() called\n");
	wdev = find_wave_device(dev, file);
	if(!wdev) {
                PWARN("Could not find wave device for file = 0x%p\n", file);
                return -ENXIO;
        }

	if (file->f_mode & FMODE_WRITE)
		poll_wait(file, &wdev->record_stream.io_wait, wait);
	if (file->f_mode & FMODE_READ)
		poll_wait(file, &wdev->play_stream.io_wait, wait);

	spin_lock_irqsave(&dev->irq_lock, flags);
	if (file->f_mode & FMODE_READ) {
		for (i=0; i < MAX_FRAGS; i++) {
			if (wdev->record_stream.buf[i].count > 0) {
				mask |= POLLIN | POLLRDNORM;
				break;
			}
		}
	}

	b_count = 0;
	if (file->f_mode & FMODE_WRITE) {
		for (i=0; i < MAX_FRAGS; i++) {
			if (wdev->play_stream.buf[i].count >= wdev->play_stream.frag_size) {
				mask |= POLLOUT | POLLWRNORM;
				break;
			}
		}
	}
	spin_unlock_irqrestore(&dev->irq_lock, flags);

	return mask;
}

static int au_audio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	au_device_t *dev = file->private_data;
	wave_device_t *wdev;
	wave_stream_t *stream;
	audio_buf_info abinfo;
	count_info cinfo;
	unsigned long int flags = 0;
	int val, i;

	// PSHOW("au_audio_ioctl, minor %d cmd 0x%x\n", dev->dev_audio, cmd);
	wdev = find_wave_device(dev, file);
	if(!wdev) {
		PWARN("Could not find wave device for file = 0x%p\n", file);
		return -ENXIO;
	}
	/* fixme: dynamic update speed, fmt, chans, etc.? */
	switch(cmd) {
	case OSS_GETVERSION:
		return put_user(SOUND_VERSION, (int *)arg);

	case SNDCTL_DSP_POST:
		if(!wdev->play_stream.wave)
			return -EINVAL;
		PSHOW("au_audio_ioctl() : DSP_POST\n");
		drain_playback(dev, wdev, (file->f_flags & O_NONBLOCK));
		stop_wave(dev, &(wdev->play_stream));
		return SUCCESS;

	case SNDCTL_DSP_SYNC:
		if (wdev->play_stream.wave) {
			PSHOW("au_audio_ioctl() : DSP_SYNC\n");
			drain_playback(dev, wdev, (file->f_flags & O_NONBLOCK));
			stop_wave(dev, &(wdev->play_stream));
		}
		if (wdev->record_stream.wave)
			stop_wave(dev, &(wdev->record_stream));
		return SUCCESS;

	case SNDCTL_DSP_SETDUPLEX:
		PSHOW("au_audio_ioctl() : DSP_SETDUPLEX\n");
		return SUCCESS;

	case SNDCTL_DSP_GETCAPS:
		PSHOW("au_audio_ioctl() : DSP_GETCAPS\n");
		return put_user(DSP_CAP_DUPLEX | DSP_CAP_REALTIME | DSP_CAP_TRIGGER | DSP_CAP_MMAP, (int *)arg);

	case SNDCTL_DSP_RESET:
		PSHOW("au_audio_ioctl() : DSP_RESET\n");
		stop_wave(dev, &(wdev->play_stream));
		stop_wave(dev, &(wdev->record_stream));
		return SUCCESS;

	case SNDCTL_DSP_SPEED:
		PSHOW("au_audio_ioctl() : DSP_SPEED\n");
		if (get_user(val, (int *)arg))
			return -EFAULT;
		wdev->play_stream.sample_rate = wdev->record_stream.sample_rate = val;
		return put_user(val, (int *)arg);

	case SNDCTL_DSP_STEREO:
		PSHOW("au_audio_ioctl() : DSP_STEREO\n");
		if (get_user(val, (int *)arg))
			return -EFAULT;
		wdev->play_stream.channel_count = wdev->record_stream.channel_count = (val ? 2 : 1);
		return put_user(val, (int *)arg);
      
	case SNDCTL_DSP_CHANNELS:
		PSHOW("au_audio_ioctl() : DSP_CHANNELS\n");
		if (get_user(val, (int *)arg))
			return -EFAULT;
		/* fixme: quad support */
		wdev->play_stream.channel_count = wdev->record_stream.channel_count = (val == 1 ? 1 : 2);
		return put_user(val, (int *)arg);
    
	case SNDCTL_DSP_GETFMTS:
		PSHOW("au_audio_ioctl() : DSP_GETFMTS\n");
		return put_user(AFMT_MU_LAW | AFMT_A_LAW | AFMT_S16_LE | AFMT_U8, (int *)arg);

	case SNDCTL_DSP_SETFMT:
		PSHOW("au_audio_ioctl() : DSP_SETFMT\n");
		if (get_user(val, (int *)arg))
			return -EFAULT;
		if(val != AFMT_QUERY) {
			set_data_format(&wdev->play_stream, val);
			set_data_format(&wdev->record_stream, val);
		}
		return put_user(wdev->play_stream.data_format, (int *)arg);

	case SNDCTL_DSP_GETTRIGGER:
		PSHOW("au_audio_ioctl() : DSP_GETTRIGGER\n");
		val = 0;
		if(wdev->record_stream.wave && !wdev->record_stream.pre_roll)
			val |= PCM_ENABLE_INPUT;
		if(wdev->play_stream.wave && !wdev->play_stream.pre_roll)
			val |= PCM_ENABLE_OUTPUT;
		return put_user(val, (int *)arg);
		
	case SNDCTL_DSP_SETTRIGGER:
		if (get_user(val, (int *)arg))
			return -EFAULT;
		if (file->f_mode & FMODE_READ) {
			if(val & PCM_ENABLE_INPUT) {
				PSHOW("DSP_SETTRIGGER Called: PCM_ENABLE_INPUT\n");
				stream = &wdev->record_stream;
				if (stream->wave && (stream->pre_roll || stream->mapped)) {
					spin_lock_irqsave(&dev->irq_lock, flags);
					if (stream->mapped && !stream->mapped_setup) {
						set_asp_format(stream);
						for (i = 0; i < stream->num_frags; i++)
							AddWaveBuffer(stream->wave, stream->buf[i].page, stream->frag_size);
						stream->mapped_setup = TRUE;
					}
					StartWave(stream->wave);
					stream->pre_roll = FALSE;
					spin_unlock_irqrestore(&dev->irq_lock, flags);
				}
			} else if (wdev->record_stream.wave) {
				if(wdev->record_stream.mapped && wdev->record_stream.mapped_setup) {
					spin_lock_irqsave(&dev->irq_lock, flags);
					StopWave(wdev->record_stream.wave);
					spin_unlock_irqrestore(&dev->irq_lock, flags);
				}
				wdev->record_stream.pre_roll = TRUE;
			}
		}
		if (file->f_mode & FMODE_WRITE) {
			if (val & PCM_ENABLE_OUTPUT) {
				PSHOW("DSP_SETTRIGGER Called: PCM_ENABLE_OUTPUT\n");
				stream = &wdev->play_stream;
				if (stream->wave && (stream->pre_roll || stream->mapped)) {
					spin_lock_irqsave(&dev->irq_lock, flags);
					if (stream->mapped && !stream->mapped_setup) {
						set_asp_format(stream);
						for (i = 0; i < stream->num_frags; i++) {
							stream->buf[i].count = stream->frag_size;
							stream->buf[i].in_use = TRUE;
							memset(stream->buf[i].page, (stream->data_format == AFMT_S16_LE) ? 0 : 0x80, stream->frag_size);
							AddWaveBuffer(stream->wave, stream->buf[i].page, stream->frag_size);
						}
						stream->mapped_setup = TRUE;	
					}
					StartWave(stream->wave);
					stream->pre_roll = FALSE;
					spin_unlock_irqrestore(&dev->irq_lock, flags);
				}
			} else if (wdev->play_stream.wave) {
				if (wdev->play_stream.mapped && wdev->play_stream.mapped_setup) {
					spin_lock_irqsave(&dev->irq_lock, flags);
					StopWave(wdev->play_stream.wave);
					spin_unlock_irqrestore(&dev->irq_lock, flags);
				}
				wdev->play_stream.pre_roll = TRUE;
			}
		}
		return SUCCESS;

	case SNDCTL_DSP_GETOSPACE:
		if(!wdev->play_stream.wave)
			return -EINVAL;
		abinfo.fragsize = wdev->play_stream.frag_size;
		abinfo.fragstotal = wdev->play_stream.num_frags;
		abinfo.bytes = 0;
		abinfo.fragments= 0;
		spin_lock_irqsave(&dev->irq_lock, flags);
		for(i = 0; i < wdev->play_stream.num_frags; i++) {
			if(!wdev->play_stream.buf[i].in_use) {
				PSHOW("Frag #%i Not in_use\n", i);
				abinfo.bytes += (wdev->play_stream.frag_size - wdev->play_stream.buf[i].count);
				if (wdev->play_stream.buf[i].count == 0) 
					++abinfo.fragments;
			} else PSHOW("Frag #%i in_use\n", i);
		}
		spin_unlock_irqrestore(&dev->irq_lock, flags);
		PSHOW("DSP_GETOSPACE called: Fragsize: %i, Fragstotal: %i, Bytes: %i, Fragments: %i\n", abinfo.fragsize, abinfo.fragstotal, abinfo.bytes, abinfo.fragments); 
		return copy_to_user((void *)arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;

	case SNDCTL_DSP_GETISPACE:
		if (!wdev->record_stream.wave)
			return -EINVAL;
		abinfo.fragsize = wdev->record_stream.frag_size;
		abinfo.fragstotal = wdev->record_stream.num_frags;
		abinfo.bytes = 0;
		abinfo.fragments= 0; 
		spin_lock_irqsave(&dev->irq_lock, flags);
		for (i = 0; i < wdev->record_stream.num_frags; i++) {
			if (!wdev->record_stream.buf[i].in_use) {
				abinfo.bytes += (wdev->record_stream.frag_size - wdev->record_stream.buf[i].count);
				if (wdev->record_stream.buf[i].count == 0)
					++abinfo.fragments;
			}
		}
		spin_unlock_irqrestore(&dev->irq_lock, flags);
		return copy_to_user((void *)arg, &abinfo, sizeof(abinfo)) ? -EFAULT : 0;

	case SNDCTL_DSP_NONBLOCK:
		PSHOW("au_audio_ioctl() : DSP_NONBLOCK\n");
		file->f_flags |= O_NONBLOCK;
		return SUCCESS;

	case SNDCTL_DSP_GETODELAY:
		val = 0;
		for (i = 0; i < wdev->play_stream.num_frags; i++)
			val += (wdev->play_stream.buf[i].count - wdev->play_stream.buf[i].xfer);
		PSHOW("DSP_GETODELAY called: Num Frags: %li, Count-XFer: %i\n", wdev->play_stream.num_frags, val);
		return put_user(val, (int *)arg);

	case SNDCTL_DSP_GETIPTR:
		stream = &wdev->record_stream;
		cinfo.bytes = stream->total_played;
		cinfo.blocks = stream->total_played / stream->frag_size;
		cinfo.ptr = (stream->frag_size * stream->fill_index) + stream->buf[stream->fill_index].xfer;
		return copy_to_user((void *)arg, &cinfo, sizeof(cinfo));

	case SNDCTL_DSP_GETOPTR:
		stream = &wdev->play_stream;
		cinfo.blocks = stream->total_played / stream->frag_size;
		cinfo.bytes = stream->total_played;
		cinfo.ptr = (stream->frag_size * stream->drain_index) + stream->buf[stream->drain_index].xfer;
		PSHOW("DSP_GETOPTR called: Bytes: %i, Blocks: %i, Ptr: %i\n", cinfo.bytes, cinfo.blocks, cinfo.ptr);
		
		return copy_to_user((void *)arg, &cinfo, sizeof(cinfo));

	case SNDCTL_DSP_GETBLKSIZE:
		PSHOW("au_audio_ioctl() : DSP_GETBLKSIZE\n");
		return put_user(wdev->play_stream.frag_size, (int *)arg);

	case SNDCTL_DSP_SETFRAGMENT:
		if (get_user(val, (int *)arg))
			return -EFAULT;
		PSHOW("DSP_SETFRAGMENT: 0x%x\n", val);
		/* fixme - because of S/G hardware we only support a fixed (4096) frag size */
		return SUCCESS;

	case SNDCTL_DSP_SUBDIVIDE:
		if (get_user(val, (int *)arg))
			return -EFAULT;		
		PSHOW("DSP_SUBDIVIDE: 0x%x\n", val);
		/* fixme */
		return SUCCESS;

	case SOUND_PCM_WRITE_FILTER:
	case SNDCTL_DSP_SETSYNCRO:
	case SOUND_PCM_READ_RATE:
	case SOUND_PCM_READ_CHANNELS:
	case SOUND_PCM_READ_BITS:
	case SOUND_PCM_READ_FILTER:
		return -EINVAL;
	}

	return au_mixer_ioctl(inode, file, cmd, arg);
}

static int au_audio_mmap(struct file *file, struct vm_area_struct *vma)
{
	au_device_t *dev = file->private_data;
	wave_device_t *wdev;
	unsigned long size;

	PSHOW("au_audio_mmap, minor = %d\n", dev->dev_audio);

	wdev = find_wave_device(dev, file);
	if (!wdev) {
		PWARN("could not find wave device for file = 0x%p\n", file);
		return -ENODEV;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
	if(vma->vm_pgoff != 0)
#else
	if (vma->vm_offset != 0)
#endif
		return -EINVAL;
	size = vma->vm_end - vma->vm_start;

	if (vma->vm_flags & VM_WRITE) {
		wave_stream_t *stream = &wdev->play_stream;
		if (!stream->wave || stream->active || stream->mapped)
			return -EINVAL;
		if (size > (PAGE_SIZE << stream->buf_order))
			return -EINVAL;
		if (remap_page_range(vma->vm_start, virt_to_phys(stream->dmabuf), size, vma->vm_page_prot))
			return -EAGAIN;
		stream->mapped = TRUE;
	}
	if (vma->vm_flags & VM_READ) {
		wave_stream_t *stream = &wdev->record_stream;
		if (!stream->wave || stream->active || stream->mapped)
			return -EINVAL;
		if (size > (PAGE_SIZE << stream->buf_order))
			return -EINVAL;
		if (remap_page_range(vma->vm_start, virt_to_phys(stream->dmabuf), size, vma->vm_page_prot))
			return -EAGAIN;
		stream->mapped = TRUE;
	}
	return SUCCESS;
}
/*
 * Open a wave audio device.
 * If FMODE_WRITE is set, open a playback stream.
 * If FMODE_READ is set, open a recording stream.
 * If both are set, full duplex is possible.
 * Multiple opens are allowed, up to the number of
 * available vortex dma channels.
 *
 * FIXME: if BLOCKING open and no dma channel is
 * free, implement a wait.
 */
static int au_audio_open(struct inode *inode, struct file *file)
{
	au_device_t *dev;
	int minor_num = MINOR(inode->i_rdev);
	int dev_type = DEVICE_TYPE(minor_num);
	int status;
	//PDEBUG("au_audio_open, minor = %d\n", minor_num);
	dev = find_device(minor_num);
	if (!dev) {
		PWARN("Could not find device for minor = %d\n", minor_num);
		return -ENODEV;
	}
	/* add new wave device and associate it with this file pointer */
	status = add_wave_device(dev, file, dev_type);
	if (status != 0) {
		PWARN("Could not allocate wave device, status = %d\n", status);
		return status;
	}
	file->private_data = (void *)dev;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	MOD_INC_USE_COUNT;
#endif
	return SUCCESS;
}

static int au_audio_release(struct inode *inode, struct file *file)
{
	au_device_t *dev = file->private_data;
	wave_device_t *wdev;
	//PDEBUG("au_audio_release, minor = %d\n", dev->dev_audio);
	wdev = find_wave_device(dev, file);
	if (!wdev) {
		PWARN("Could not find wave device for file = 0x%p\n", file);
		return -ENODEV;
	}
  	if (wdev->play_stream.wave)
		drain_playback(dev, wdev, (file->f_flags & O_NONBLOCK));
	if (wdev->record_stream.wave)
		dev->record_in_use = FALSE;
  	/* kill streams */
	stop_wave(dev, &(wdev->play_stream));
	stop_wave(dev, &(wdev->record_stream));
	remove_wave_device(dev, wdev);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	MOD_DEC_USE_COUNT;
#endif
	return SUCCESS;
}

static void drain_playback(au_device_t *dev, wave_device_t *wdev, int nonblock)
{
	unsigned long flags = 0;
	mem_buf_t *mb = &(wdev->play_stream.buf[wdev->play_stream.fill_index]);

	PSHOW("drain_playback() : entry\n");
	if (wdev->play_stream.mapped) {
		PSHOW("drain_playback() : exit was 1 (stream mapped)\n");
		return;
	}
	if (!mb->in_use && mb->count) {
		mb->in_use=TRUE;
		mb->xfer=0;
		spin_lock_irqsave(&dev->irq_lock,flags);
		AddWaveBuffer(wdev->play_stream.wave, mb->page, mb->count);
		spin_unlock_irqrestore(&dev->irq_lock,flags);
	}

 	while (wdev->play_stream.total_played < wdev->play_stream.total_sent) {
		PSHOW("drain_playback() : loop \n");
		if (!wdev->play_stream.active) {
			spin_lock_irqsave(&dev->irq_lock, flags);
			StartWave(wdev->play_stream.wave);
			wdev->play_stream.active = TRUE;
			spin_unlock_irqrestore(&dev->irq_lock, flags);
		}
		if(nonblock)
			break;
		interruptible_sleep_on(&wdev->play_stream.io_wait);
		if (signal_pending(current)) {
			PSHOW("drain_playback() : exit was 1 (app killed) \n");
			break;
		}
	}
	PSHOW("drain_playback() : exit was 2 (normal) \n");
}

static int add_wave_device(au_device_t *dev, struct file *file, int dev_type)
{
	wave_device_t *wdev;
	int status;
	unsigned long flags = 0;
	int enable_playback = file->f_mode & FMODE_WRITE;
	int enable_record = file->f_mode & FMODE_READ;
	ASPWAVEFORMAT fmt = { 8, 1, 8000, ASPFMTLINEAR8 }; /* arbitrary format, reset by read/write */

	/* only allow one record stream */
	if (enable_record && dev->record_in_use)
		return -EBUSY;
	wdev = (wave_device_t *)kmalloc(sizeof(wave_device_t), GFP_KERNEL);
	if(!wdev)
		return -ENOMEM;
	memset(wdev, 0, sizeof(wave_device_t));
	/* set default data format based on device type */
	wdev->play_stream.sample_rate = 8000;
	wdev->play_stream.channel_count = 1;
	wdev->record_stream.sample_rate = 8000;
	wdev->record_stream.channel_count = 1;
	switch(dev_type) {
		case DEV_AUDIO:
		default:
			wdev->play_stream.data_format = AFMT_U8;
			wdev->record_stream.data_format = AFMT_U8;
			break;
		case DEV_SUNDSP:
			wdev->play_stream.data_format = AFMT_MU_LAW;
			wdev->record_stream.data_format = AFMT_MU_LAW;
			break;
	}

	wdev->fp = file;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
	init_waitqueue_head(&wdev->play_stream.io_wait);
	init_waitqueue_head(&wdev->record_stream.io_wait);
#else
	init_waitqueue(&wdev->play_stream.io_wait);
	init_waitqueue(&wdev->record_stream.io_wait);
#endif

	/* fixme: frag size should be dynamic and calculated based on stream format */
	wdev->play_stream.frag_size = DEF_FRAG_SIZE;
	wdev->play_stream.num_frags = DEF_NR_FRAGS;
	wdev->record_stream.frag_size = DEF_FRAG_SIZE;
	wdev->record_stream.num_frags = DEF_NR_FRAGS;

	if (enable_playback) {
		if(get_frags(&(wdev->play_stream), wdev->play_stream.num_frags) != SUCCESS) {
			PWARN("Could not get playback stream pages.\n");
			free_wave_device(dev, wdev);
			return -ENOMEM;
		}
		status = CreateWaveBuffer(dev->core, &fmt, ASPDIRPLAY, &wdev->play_stream.wave);
		if (status != 0) {
			PWARN("could not create play stream wave buffer, status=%d\n", status);
			free_wave_device(dev, wdev);
			return -EBUSY;	// ran out of dma channels
		}
	}
	if (enable_record) {
		if(get_frags(&(wdev->record_stream), wdev->record_stream.num_frags) != SUCCESS) {
			PWARN("Could not get record stream pages\n");
			free_wave_device(dev, wdev);
			return -ENOMEM;
		}
		status = CreateWaveBuffer(dev->core, &fmt, ASPDIRRECORD, &wdev->record_stream.wave);
		if(status != 0) {
			PWARN("Could not create record stream wave buffer, status=%d\n", status);
			free_wave_device(dev, wdev);
			return -ENOMEM;
		}
		dev->record_in_use = TRUE;
	}
	spin_lock_irqsave(&dev->irq_lock, flags);
	/* link to end of wave device list */
	if (!dev->wave_head)
		dev->wave_head = dev->wave_tail = wdev;
	else {
		dev->wave_tail->next = wdev;
		wdev->prev = dev->wave_tail;
		dev->wave_tail = wdev;
	}
	spin_unlock_irqrestore(&dev->irq_lock, flags);
	return SUCCESS;  
}

static int get_frags(wave_stream_t *stream, unsigned long count)
{
	int order = 2; // fixme: base on count
	unsigned long page_addr;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
	struct page *map, *mapend;
#else
	unsigned long map, mapend;
#endif
	int i;

	stream->dmabuf = (void *)__get_free_pages(GFP_KERNEL, order);
	if (!stream->dmabuf)
		return -ENOMEM;
	stream->buf_order = order;
	/* now mark the pages as reserved; otherwise remap_page_range doesn't do what we want */
	/* this is from the es1371 driver */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
	mapend = virt_to_page(stream->dmabuf + (PAGE_SIZE << order) - 1);
	for(map = virt_to_page(stream->dmabuf); map <= mapend; map++)
		mem_map_reserve(map);
#else
	mapend = MAP_NR(stream->dmabuf + (PAGE_SIZE << order) - 1);
	for(map = MAP_NR(stream->dmabuf); map <= mapend; map++)
		set_bit(PG_reserved, &mem_map[map].flags);
#endif
	page_addr = (unsigned long)stream->dmabuf;
	for(i = 0; i < count; i++) {
		stream->buf[i].page = (void *)page_addr;
		stream->buf[i].in_use = FALSE;
		page_addr += DEF_FRAG_SIZE;
	}
	return SUCCESS;
}

static void free_wave_device(au_device_t *dev, wave_device_t *wdev)
{
	unsigned long flags = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
	struct page *map, *mapend;
#else
	unsigned long map, mapend;
#endif

	if (wdev->play_stream.dmabuf) {
		/* undo marking the pages as reserved */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
		mapend = virt_to_page(wdev->play_stream.dmabuf + (PAGE_SIZE << wdev->play_stream.buf_order) - 1);
		for (map = virt_to_page(wdev->play_stream.dmabuf); map <= mapend; map++)
			mem_map_unreserve(map);
#else
		mapend = MAP_NR(wdev->play_stream.dmabuf + (PAGE_SIZE << wdev->play_stream.buf_order) - 1);
		for(map = MAP_NR(wdev->play_stream.dmabuf); map <= mapend; map++)
			clear_bit(PG_reserved, &mem_map[map].flags);	
#endif
		free_pages((unsigned long)wdev->play_stream.dmabuf, wdev->play_stream.buf_order);
	}
	if (wdev->record_stream.dmabuf) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
		mapend = virt_to_page(wdev->record_stream.dmabuf + (1 << wdev->record_stream.buf_order));
		for (map = virt_to_page(wdev->record_stream.dmabuf); map < mapend; map++)
			mem_map_unreserve(map);
#else
		mapend = MAP_NR(wdev->record_stream.dmabuf + (PAGE_SIZE << wdev->record_stream.buf_order) - 1);
		for (map = MAP_NR(wdev->record_stream.dmabuf); map <= mapend; map++)
			clear_bit(PG_reserved, &mem_map[map].flags);	
#endif
		free_pages((unsigned long)wdev->record_stream.dmabuf, wdev->record_stream.buf_order);
	}
	if (wdev->play_stream.wave) {
		spin_lock_irqsave(&dev->irq_lock, flags);
		ReleaseWave(wdev->play_stream.wave);
		spin_unlock_irqrestore(&dev->irq_lock, flags);
	}
	if (wdev->record_stream.wave) {
		spin_lock_irqsave(&dev->irq_lock, flags);
		ReleaseWave(wdev->record_stream.wave);
		spin_unlock_irqrestore(&dev->irq_lock, flags);
	}
	kfree(wdev);
}

static wave_device_t* find_wave_device(au_device_t *dev, struct file *file)
{
	unsigned long flags = 0;
	wave_device_t *wdev;

	/* could use reader/writer spinlocks */
	spin_lock_irqsave(&dev->irq_lock, flags);

	wdev = dev->wave_head;
	while (wdev) {
		if (wdev->fp == file)
			break;
		wdev = wdev->next;
	}
	spin_unlock_irqrestore(&dev->irq_lock, flags);

	return wdev;
}

static void remove_wave_device(au_device_t *dev, wave_device_t *wdev)
{
	unsigned long flags = 0;
	spin_lock_irqsave(&dev->irq_lock, flags);
	/* unlink the wave device */
	if(wdev == dev->wave_head) {
		dev->wave_head = wdev->next;
		if(dev->wave_head)
			dev->wave_head->prev = NULL;
	} else {
		wdev->prev->next = wdev->next;
		if(wdev->next)
			wdev->next->prev = wdev->prev;
		else
			dev->wave_tail = wdev->prev;
	}
	spin_unlock_irqrestore(&dev->irq_lock, flags);
	free_wave_device(dev, wdev);
}
/*
 * If the passed format is not valid, the stream's format
 * is not changed.
 */
static void set_data_format(wave_stream_t *stream, int val)
{
	switch(val) {
		case AFMT_MU_LAW:
		case AFMT_A_LAW:
		case AFMT_S16_LE:
		case AFMT_U8:
			stream->data_format = val;
			break;
		case AFMT_S16_BE:
		case AFMT_U16_LE:
		case AFMT_U16_BE:
		case AFMT_IMA_ADPCM:
		case AFMT_S8:
		/* fixme: do the conversion */
		default:
/*			PWARN("Unsupported data format: %d\n", val);*/
	}
}

static void set_asp_format(wave_stream_t *stream)
{
	ASPWAVEFORMAT fmt;

	switch(stream->data_format) {
		case AFMT_S16_LE:
		default:
			fmt.eEncoding = ASPFMTLINEAR16;
			fmt.wBitWidth = 16;
			break;
		case AFMT_U8:
			fmt.eEncoding = ASPFMTLINEAR8;
			fmt.wBitWidth = 8;
			break;
		case AFMT_A_LAW:
			fmt.eEncoding = ASPFMTALAW;
			fmt.wBitWidth = 8;
			break;
		case AFMT_MU_LAW:
			fmt.eEncoding = ASPFMTULAW;
			fmt.wBitWidth = 8;
			break;
	}
	fmt.wChannels = stream->channel_count ? stream->channel_count : 1;
	fmt.dwSampleRate = stream->sample_rate ? stream->sample_rate : 8000;
	//PDEBUG("bw %d ch %d sr %d en %d\n", fmt.wBitWidth, fmt.wChannels, (int)fmt.dwSampleRate, fmt.eEncoding);
	SetWaveFormat(stream->wave, &fmt);
}

static void stop_wave(au_device_t *dev, wave_stream_t *stream)
{
	int i;
	unsigned long flags = 0;

	if (stream->wave) {
		spin_lock_irqsave(&dev->irq_lock, flags);
		StopWave(stream->wave);
		spin_unlock_irqrestore(&dev->irq_lock, flags);
		}
	stream->active = stream->pre_roll = FALSE;
	stream->total_played = stream->total_sent = 0;
	stream->fill_index = stream->drain_index = 0;
	for (i = 0; i < MAX_FRAGS; i++) {
		stream->buf[i].count = stream->buf[i].xfer = 0;
		stream->buf[i].in_use = FALSE;
	}
}

static unsigned long receive_mem(au_device_t *dev, wave_stream_t *stream, char *mem, unsigned long max_size)
{
	unsigned long flags = 0;
	unsigned long received = 0;
	unsigned long count;
	int i;

        /*
         * Add the first buffers to an inactive stream.
         */
        if (!stream->active) {
		spin_lock_irqsave(&dev->irq_lock, flags);
                for (i = 0; i < stream->num_frags; i++) {
                        stream->buf[i].count = stream->frag_size;
                        stream->buf[i].in_use = TRUE;
                        AddWaveBuffer(stream->wave, stream->buf[i].page, stream->frag_size);
                }
		spin_unlock_irqrestore(&dev->irq_lock, flags);
                return 0; /* no memory received yet */
        }

	while ((received < max_size) && !stream->buf[stream->drain_index].in_use) {
		count = max_size < stream->frag_size ? max_size : stream->frag_size;
		copy_to_user(mem, stream->buf[stream->drain_index].page, count);
		/*
		 * Add this buffer back.
		 */
		stream->buf[stream->drain_index].count = stream->frag_size;
		stream->buf[stream->drain_index].xfer = 0;
		stream->buf[stream->drain_index].in_use = TRUE;
		spin_lock_irqsave(&dev->irq_lock, flags);
		AddWaveBuffer(stream->wave, stream->buf[stream->drain_index].page, stream->frag_size);
		spin_unlock_irqrestore(&dev->irq_lock, flags);

		stream->drain_index++;
		if (stream->drain_index == stream->num_frags)
			stream->drain_index = 0;
		received += count;
		mem += count;
	}
	return received;
}

static unsigned long send_mem(au_device_t *dev, wave_stream_t *stream, const char *mem, unsigned long size)
{
	unsigned long flags = 0;
	unsigned long sent = 0;
	unsigned long count;
	mem_buf_t *mb = &stream->buf[stream->fill_index];
	/*
	 * Put as much as we can into each unused buffer.
	 */
	PSHOW("send_mem() : entry\n");
	PSHOW("send_mem() : size is %li\n", size);
	while (size && !mb->in_use) {
		PSHOW("send_mem() : loop\n");
		count = size < (stream->frag_size - mb->count) ? size : stream->frag_size - mb->count;
		copy_from_user(mb->page + mb->count, mem, count);
		mb->count += count;
		PSHOW("send_mem() : count     = %li\n", count);
		PSHOW("send_mem() : mb_count  = %li\n", mb->count);
		PSHOW("send_mem() : fillindex = %i\n",  stream->fill_index);
		PSHOW("send_mem() : numfrags  = %li\n", stream->num_frags);
		/*
		 * Send buffer to the dma engine if full.
		 */
		if (mb->count >= stream->frag_size) {
			mb->in_use = TRUE;
			mb->xfer = 0;
		/* XXX: calling PSHOW with a spinlock held is probably not a great idea */
			spin_lock_irqsave(&dev->irq_lock, flags);
			if(AddWaveBuffer(stream->wave, mb->page, mb->count)) {
				count = 0;
				PSHOW("send_mem() : AddWaveBuffer() failed\n");
			}
			spin_unlock_irqrestore(&dev->irq_lock, flags);
			PSHOW("send_mem() : AddWaveBuffer() success for %li bytes.\n", mb->count);
			stream->fill_index++;
			if(stream->fill_index >= stream->num_frags)
				stream->fill_index = 0;
			mb = &stream->buf[stream->fill_index];
		}
		PSHOW("send_mem() : fillindex = %i\n",  stream->fill_index);
		PSHOW("send_mem() : numfrags  = %li\n", stream->num_frags);
		size  -= count;
		sent  += count;
		mem   += count;
	}
	return sent;
}

/* au_handle_timer is called with the spinlock held */
void au_handle_timer(au_device_t *dev)
{
	wave_device_t *wdev = dev->wave_head;
	wave_stream_t *stream;
	mem_buf_t *cur;
	unsigned long count, remaining;

	while(wdev) {
		if (wdev->play_stream.active || wdev->play_stream.mapped) {
			stream = &wdev->play_stream;
			count = GetWaveBytesPlayed(stream->wave);
			if((!count)&&(stream->total_played < stream->total_sent)) {
				PSHOW("DLDF activated! c = 0 p = %li t = %li d = %li\n", stream->total_played, stream->total_sent, stream->total_sent - stream->total_played);
				stream->total_played = stream->total_sent;				
			}
			if(!count)
				wake_up_interruptible(&stream->io_wait);
			stream->total_played += count;
			PSHOW("au_handle_timer() : pl = %li tt = %li d = %li\n", stream->total_played, stream->total_sent, stream->total_sent - stream->total_played);
			while(count) {
				cur = &(stream->buf[stream->drain_index]);
//				assert cur->in_use
//				stream->total_played += count;					// [MARK]
				remaining = cur->count - cur->xfer;
				if(count < remaining) {
					cur->xfer += count;
					count = 0;
				} else {
					stream->drain_index++;
					if(stream->drain_index == stream->num_frags)
						stream->drain_index = 0;
					if(stream->mapped) {
						cur->xfer = 0;
						AddWaveBuffer(stream->wave, cur->page, cur->count);
					} else {
						cur->xfer = cur->count;
						cur->count = 0;
						cur->in_use = FALSE;
						wake_up_interruptible(&stream->io_wait);
					}
					count -= remaining;
				}
			}
		}

		if(wdev->record_stream.active || wdev->record_stream.mapped) {
			stream = &wdev->record_stream;
			count = GetWaveBytesPlayed(stream->wave);
			stream->total_played += count;
			while(count) {
				cur = &(stream->buf[stream->fill_index]);
				/* assert cur->in_use */
//				stream->total_played += count;
				remaining = cur->count - cur->xfer;
				if (count < remaining) {
					cur->xfer += count;
					count = 0;
				} else {
					stream->fill_index++;
					if(stream->fill_index == stream->num_frags)
						stream->fill_index = 0;
					if(stream->mapped) {
						cur->xfer = 0;
						AddWaveBuffer(stream->wave, cur->page, cur->count);
					} else {
						cur->xfer = cur->count;
						cur->count = 0;
						cur->in_use = FALSE;
						wake_up_interruptible(&stream->io_wait);
					}
					count -= remaining;
				}
			}
		}
		wdev = wdev->next;
	}
}
