/*
 * au_midi.c
 *
 * Aureal Vortex AU88xx audio driver module
 * MIDI Audio
 *
 * Copyright (c) 1999 Aureal Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Some of this code is based on the es1371 driver, which is
 * Copyright (C) 1998  Thomas Sailer (sailer@ife.ee.ethz.ch)
 * 
 * Revision History:
 * 
 * 19-08-2000 Nikos Kalogridis
 *      Split vortex.c into multiple files for easier maintenance
 *
 * 24-08-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Reformatted Code
 * 
 */

#define __NO_VERSION__
#include "au_vortex.h"
#include <linux/version.h>

/* Midi file operations */

static int au_midi_open(struct inode *inode, struct file *file);
static int au_midi_release(struct inode *inode, struct file *file);
static int au_midi_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static ssize_t au_midi_read(struct file *file, char *buffer, size_t count, loff_t *ppos);
static ssize_t au_midi_write(struct file *file, const char *buffer, size_t count, loff_t *ppos);
static unsigned int au_midi_poll(struct file *file, struct poll_table_struct *wait);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
static struct file_operations midi_fops =
{
	owner:		THIS_MODULE,
	llseek:		au_llseek,
	read:		au_midi_read,
	write:		au_midi_write,
	poll:		au_midi_poll,
	ioctl:		au_midi_ioctl,
	open:		au_midi_open,
	release:	au_midi_release,
};
#else
static struct file_operations midi_fops = {
        au_llseek,
        au_midi_read,
        au_midi_write,   
        NULL,                   /* readdir */
        au_midi_poll,
        au_midi_ioctl,
        NULL,                   /* mmap */
        au_midi_open,
        NULL,                   /* flush */
        au_midi_release,
        NULL,                   /* fsync */
        NULL,                   /* fasync */
        NULL,                   /* check media change */
        NULL,                   /* revalidate */
        NULL,                   /* lock */
};
#endif

/********************************/
/* Midi support			*/
/********************************/

int au_midi_register(au_device_t *dev)
{
	unsigned char clock_divider = 0x61;
	unsigned char midi_mode = 0;
	int mpu_mode = 0x1;
	unsigned long midi_port = 3; // 330-331
	int status;
	
	PDEBUG("au_midi_register\n");
	/* fixme - make legacy mpu port address selectable? */
	EnableHardCodedMPU401Port(dev->core, TRUE, midi_port);
	if(!Mpu401UartInit(dev->core, clock_divider, midi_mode, mpu_mode))
		PWARN("No ACK from midi port!\n");
	
	/* register midi device with soundcore */
	status = dev->dev_midi = register_sound_midi(&midi_fops, -1);
	return (status < 0) ? status : 0;
}

int au_midi_unregister(au_device_t *dev)
{
	PDEBUG("au_midi_unregister\n");
	unregister_sound_midi(dev->dev_midi);
	return 0;
}

static int au_midi_open(struct inode *inode, struct file *file)
{
	au_device_t *dev;
	int minor_num = MINOR(inode->i_rdev);
	int enable_output = file->f_mode & FMODE_WRITE;
	int enable_input = file->f_mode & FMODE_READ;
	unsigned long flags;

	//PDEBUG("au_midi_open, minor = %d\n", minor_num);

	dev = find_device(minor_num);
	if (!dev) {
		PWARN("could not find device for minor = %d\n", minor_num);
		return -ENODEV;
	}

	/* FIXME: if BLOCKING open and in use, implement a wait. */
	if (enable_output) {
		if (dev->midiout_in_use)
			return -EBUSY;
		dev->midiout_in_use = TRUE;
	}
	if (enable_input) {
		if (dev->midiin_in_use)
			return -EBUSY;
		dev->midiin_in_use = TRUE;
	}

	file->private_data = (void *)dev;

  	spin_lock_irqsave(&dev->irq_lock, flags);
	if (enable_output)
		dev->midi.ord = dev->midi.owr = dev->midi.ocnt = 0;
	if (enable_input)
		dev->midi.ird = dev->midi.iwr = dev->midi.icnt = 0;
	au_handle_midi(dev);
  	EnableMPU401Interrupt(dev->core, TRUE);
  	spin_unlock_irqrestore(&dev->irq_lock, flags);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	MOD_INC_USE_COUNT;
#endif
	return SUCCESS;
}

static int au_midi_release(struct inode *inode, struct file *file)
{
	au_device_t *dev = file->private_data;

	//PDEBUG("au_midi_release, minor = %d\n", dev->dev_midi);

	EnableMPU401Interrupt(dev->core, FALSE);
	dev->midiin_in_use = dev->midiout_in_use = FALSE;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	MOD_DEC_USE_COUNT;
#endif
	return SUCCESS;
}

static ssize_t au_midi_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	au_device_t *dev = file->private_data;
	ssize_t ret = 0;
	unsigned long flags;
	unsigned ptr;
	int cnt;

	//PDEBUG("au_midi_read, count=%d\n", count);

	if (ppos != &file->f_pos)
		return -ESPIPE;

	if (!access_ok(VERIFY_WRITE, buffer, count))
		return -EFAULT;

	while (count > 0) {
		spin_lock_irqsave(&dev->irq_lock, flags);
		ptr = dev->midi.ird;
		cnt = MIDIINBUF - ptr;
		if (dev->midi.icnt < cnt)
			cnt = dev->midi.icnt;
		spin_unlock_irqrestore(&dev->irq_lock, flags);
		if (cnt > count)
			cnt = count;
		if (cnt <= 0) {
			if (file->f_flags & O_NONBLOCK)
				return ret ? ret : -EAGAIN;
			interruptible_sleep_on(&dev->midi.iwait);
			if (signal_pending(current))
				return ret ? ret : -ERESTARTSYS;
			continue;
		}
		if (copy_to_user(buffer, dev->midi.ibuf + ptr, cnt))
			return ret ? ret : -EFAULT;
		ptr = (ptr + cnt) % MIDIINBUF;
		spin_lock_irqsave(&dev->irq_lock, flags);
		dev->midi.ird = ptr;
		dev->midi.icnt -= cnt;
		spin_unlock_irqrestore(&dev->irq_lock, flags);
		count -= cnt;
		buffer += cnt;
		ret += cnt;
	}
	return ret;
}

static ssize_t au_midi_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
	au_device_t *dev = file->private_data;
	ssize_t ret = 0;
	unsigned long flags;
	unsigned ptr;
	int cnt;

	//PDEBUG("au_midi_write, count=%d\n", count);

	if (ppos != &file->f_pos)
		return -ESPIPE;

	if (!access_ok(VERIFY_READ, buffer, count))
		return -EFAULT;

	while (count > 0) {
		spin_lock_irqsave(&dev->irq_lock, flags);
		ptr = dev->midi.owr;
		cnt = MIDIOUTBUF - ptr;
		if (dev->midi.ocnt + cnt > MIDIOUTBUF)
			cnt = MIDIOUTBUF - dev->midi.ocnt;
		if (cnt <= 0)
			au_handle_midi(dev);
		spin_unlock_irqrestore(&dev->irq_lock, flags);
		if (cnt > count)
			cnt = count;
		if (cnt <= 0) {
			if (file->f_flags & O_NONBLOCK)
				return ret ? ret : -EAGAIN;
			interruptible_sleep_on(&dev->midi.owait);
			if (signal_pending(current))
				return ret ? ret : -ERESTARTSYS;
			continue;
		}
		if (copy_from_user(dev->midi.obuf + ptr, buffer, cnt))
			return ret ? ret : -EFAULT;
		ptr = (ptr + cnt) % MIDIOUTBUF;
		spin_lock_irqsave(&dev->irq_lock, flags);
		dev->midi.owr = ptr;
		dev->midi.ocnt += cnt;
		spin_unlock_irqrestore(&dev->irq_lock, flags);
		count -= cnt;
		buffer += cnt;
		ret += cnt;
		spin_lock_irqsave(&dev->irq_lock, flags);
		au_handle_midi(dev);
		spin_unlock_irqrestore(&dev->irq_lock, flags);
	}
	return ret;
}

static unsigned int au_midi_poll(struct file *file, struct poll_table_struct *wait)
{
	au_device_t *dev = file->private_data;
	unsigned long flags;
	unsigned int mask = 0;

	if (file->f_mode & FMODE_WRITE)
		poll_wait(file, &dev->midi.owait, wait);
	if (file->f_mode & FMODE_READ)
		poll_wait(file, &dev->midi.iwait, wait);
	spin_lock_irqsave(&dev->irq_lock, flags);
	if (file->f_mode & FMODE_READ) {
		if (dev->midi.icnt > 0)
			mask |= POLLIN | POLLRDNORM;
	}
	if (file->f_mode & FMODE_WRITE) {
		if (dev->midi.ocnt < MIDIOUTBUF)
			mask |= POLLOUT | POLLWRNORM;
	}
	spin_unlock_irqrestore(&dev->irq_lock, flags);
	return mask;
}

static int au_midi_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	/* fixme */
	return SUCCESS;
}

void au_handle_midi(au_device_t *dev)
{
	unsigned char ch;
	int wake;

	//PDEBUG("au_handle_midi\n");

	if (!(dev->midiin_in_use || dev->midiout_in_use))
		return;
	wake = FALSE;
	while ((Mpu401UartReadStatus(dev->core) & MPU401_MIDIVAL) == 0) {
		ch = Mpu401UartReadData(dev->core);
		if (dev->midi.icnt < MIDIINBUF) {
			dev->midi.ibuf[dev->midi.iwr] = ch;
			dev->midi.iwr = (dev->midi.iwr + 1) % MIDIINBUF;
			dev->midi.icnt++;
		}
		wake = TRUE;
	}
  	if (wake)
  		wake_up_interruptible(&dev->midi.iwait);

	wake = FALSE;
	while (dev->midi.ocnt > 0 && ((Mpu401UartReadStatus(dev->core) & MPU401_CMDOK) == 0)) {
		Mpu401UartWriteData(dev->core, dev->midi.obuf[dev->midi.ord]);
		dev->midi.ord = (dev->midi.ord + 1) % MIDIOUTBUF;
		dev->midi.ocnt--;
		if (dev->midi.ocnt < MIDIOUTBUF-16)
			wake = TRUE;
	}
  	if (wake)
  		wake_up_interruptible(&dev->midi.owait);
}
