/* 
 * au_mixer.c
 * 
 * Aureal Vortex AU88xx audio driver module
 * Mixer Functions
 * 
 * Copyright (c) 1999 Aureal Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Some of this code is based on the es1371 driver, which is
 * Copyright (C) 1998  Thomas Sailer (sailer@ife.ee.ethz.ch)
 * 
 * Revision History:
 *
 * 15-07-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Mixer correctly reports the name of the sound chip
 *
 * 19-08-2000 Nikos Kalogridis
 *      Split vortex.c into multiple files for easier maintenance
 *
 * 20-08-2000 Nikos Kalogridis
 *      Exposed the au_mixer_ioctl in the au_vortex.h as it is needed
 *      by the au_audio.c file as well. Small bug produced by file
 *      split
 *
 * 24-08-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Reformatted Code
 *
 */

#define __NO_VERSION__
#include "au_vortex.h"
#include <linux/version.h>

static int au_mixer_open(struct inode *inode, struct file *file);
static int au_mixer_release(struct inode *inode, struct file *file);
extern void SpdifSetSampleRate(ClassAsp4Core *pCore, unsigned long sample_rate);
extern void SpdifGetSampleRate(ClassAsp4Core *pCore, unsigned long *pSample_rate);



/* Mixer file operations */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
static struct file_operations mixer_fops =
{
	owner:		THIS_MODULE,
	llseek:		au_llseek,
	ioctl:		au_mixer_ioctl,
	open:		au_mixer_open,
	release:	au_mixer_release,
};
#else
static struct file_operations mixer_fops = {
        au_llseek,
        NULL,                   /* read */ 
        NULL,                   /* write */
        NULL,                   /* readdir */
        NULL,                   /* poll */  
        au_mixer_ioctl,
        NULL,                   /* mmap */
        au_mixer_open,
        NULL,                   /* flush */
        au_mixer_release,
        NULL,                   /* fsync */
        NULL,                   /* fasync */
        NULL,                   /* check media change */
        NULL,                   /* revalidate */
        NULL,                   /* lock */  
};
#endif

static void write_mixer(au_device_t *dev, unsigned int chan, int val);
static void init_mixer(au_device_t *dev);
static void set_rec_src(au_device_t *dev, unsigned int val);

/********************************/
/* Mixer support		*/
/********************************/

int au_mixer_register(au_device_t *dev)
{
	int status;
	
	/* register mixer device with soundcore */
	PDEBUG("au_mixer_register\n");
	status = dev->dev_mixer = register_sound_mixer(&mixer_fops, -1);
	/* also init the mixer */
	init_mixer(dev);
	return (status < 0) ? status : 0;
}

int au_mixer_unregister(au_device_t *dev)
{
	PDEBUG("au_sndstat_unregister\n");
	unregister_sound_mixer(dev->dev_mixer);
	return 0;
}

int au_mixer_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	au_device_t *dev = file->private_data;
	unsigned long temp;
	int val, i;

	//PDEBUG("au_mixer_ioctl, minor %d cmd 0x%x\n", dev->dev_mixer, cmd);

	if (cmd == OSS_GETVERSION) {
		return put_user(SOUND_VERSION, (int *)arg);
	}
	if (cmd == SOUND_MIXER_INFO) {
		mixer_info info;
                strncpy(info.id, sound_proc_get_chip(dev->dev_id), sizeof(info.id));
                strncpy(info.name, "Aureal ", sizeof(info.name));
		strncat(info.name, info.id, sizeof(info.name));
		info.modify_counter = dev->mix_modcnt;
		if (copy_to_user((void *)arg, &info, sizeof(info)))
			return -EFAULT;
		return SUCCESS;
	}
	if (cmd == SOUND_OLD_MIXER_INFO) {
		_old_mixer_info info;
		strncpy(info.id, sound_proc_get_chip(dev->dev_id), sizeof(info.id));
		strncpy(info.name, "Aureal ", sizeof(info.name));  
		strncat(info.name, info.id, sizeof(info.name));
		if (copy_to_user((void *)arg, &info, sizeof(info)))
			return -EFAULT;
		return SUCCESS;
	}

	if (cmd == SOUND_MIXER_PRIVATE1) {      /* aka AUREAL_MIXER_SPDIF */
		if (get_user(val, (int *)arg))
			return -EFAULT;
		PDEBUG("PRIVATE1 (SPDIF): %d\n", val);
		if (val == 0) { /* query rate */
			temp = 44100;
			SpdifGetSampleRate(dev->core, &temp);
			val = temp;
			return put_user(val, (int *)arg);
		} else {
			SpdifSetSampleRate(dev->core, (unsigned long)val);
			return SUCCESS;
		}
	}

	if (_IOC_TYPE(cmd) != 'M' || _IOC_SIZE(cmd) != sizeof(int)) {
		//fixme I get a 0x5401 on RedHat 6.1
		//PWARN("invalid mixer read ioctl 0x%x\n", cmd);
		return -EINVAL;
	}

	if (_IOC_DIR(cmd) == _IOC_READ) {
		switch (_IOC_NR(cmd)) {
    
		case SOUND_MIXER_RECSRC:
			return put_user(dev->mix_recsrc, (int *)arg);

		case SOUND_MIXER_DEVMASK:
			/* fixme: get codec caps and add tone, headphone... */
			return put_user(SOUND_MASK_LINE | SOUND_MASK_CD | SOUND_MASK_VIDEO |
					SOUND_MASK_LINE1 | SOUND_MASK_PCM | SOUND_MASK_VOLUME |
					SOUND_MASK_OGAIN | SOUND_MASK_PHONEIN | SOUND_MASK_SPEAKER |
					SOUND_MASK_MIC | SOUND_MASK_RECLEV, (int *)arg);

		case SOUND_MIXER_RECMASK:
			return put_user(SOUND_MASK_MIC | SOUND_MASK_CD | SOUND_MASK_VIDEO | SOUND_MASK_LINE1 |
					SOUND_MASK_LINE | SOUND_MASK_VOLUME | SOUND_MASK_PHONEOUT |
					SOUND_MASK_PHONEIN, (int *)arg);

		case SOUND_MIXER_STEREODEVS:
			return put_user(SOUND_MASK_LINE | SOUND_MASK_CD | SOUND_MASK_VIDEO |
					SOUND_MASK_LINE1 | SOUND_MASK_PCM | SOUND_MASK_VOLUME |
					SOUND_MASK_PHONEOUT | SOUND_MASK_RECLEV, (int *)arg);

		case SOUND_MIXER_CAPS:
			return put_user(SOUND_CAP_EXCL_INPUT, (int *)arg);

		default:
			i = _IOC_NR(cmd);
			if (i >= SOUND_MIXER_NRDEVICES)
				return -EINVAL;
			return put_user(dev->mix_vol[i], (int *)arg);
		}
	}

	if (_IOC_DIR(cmd) != (_IOC_READ|_IOC_WRITE)) {
		PWARN("invalid mixer write ioctl 0x%x\n", cmd);
		return -EINVAL;
	}

	dev->mix_modcnt++;
	switch (_IOC_NR(cmd)) {

	case SOUND_MIXER_RECSRC:
		if (get_user(dev->mix_recsrc, (int *)arg))
			return -EFAULT;
		set_rec_src(dev, dev->mix_recsrc);
		return SUCCESS;

	default:
		i = _IOC_NR(cmd);
		if (i >= SOUND_MIXER_NRDEVICES)
			return -EINVAL;
		if (get_user(val, (int *)arg))
			return -EFAULT;
		write_mixer(dev, (unsigned)i, val);
		dev->mix_vol[i] = val;
		return put_user(val, (int *)arg);
	}
}

static void write_mixer(au_device_t *dev, unsigned int chan, int val)
{
	unsigned int left, right;
	unsigned int boost = 0;
	short reg;

	left = val & 0xff;
	right = (val >> 8) & 0xff;
	if (left > 100)
		left = 100;
	if (right > 100)
		right = 100;

	switch (chan) {
	case SOUND_MIXER_VOLUME:
		if (left < 7 && right < 7) {
			WriteCodec(dev->core, AC97REG_MASTER, 0x8000);
			break;
		}
		if (left < 7)
			left = 7;
		if (right < 7)
			right = 7;
		/* codec scales 0x1f (-46db) to 0 (loudest) */
		WriteCodec(dev->core, AC97REG_MASTER, (((100 - left) / 3) << 8) | ((100 - right) / 3));
		break;
	case SOUND_MIXER_PCM:
		if (dev->dev_id == DEV_AU8820) {
			// Must update codec pcm register on 8820
			if (left < 7 && right < 7) {
				WriteCodec(dev->core, AC97REG_PCM, 0x8000);
				break;
			} else {
				WriteCodec(dev->core, AC97REG_PCM, 0x0606);
			}
		}
		/* pcm scales 0 to 65535 */
		WriteMixer(dev->core, ASP_MIX_INPUT_WAVE_AT_PCM, (left * 655), (right * 655));
		break;
	case SOUND_MIXER_LINE:
	case SOUND_MIXER_VIDEO:
	case SOUND_MIXER_LINE1:
	case SOUND_MIXER_CD:
		switch (chan) {
		case SOUND_MIXER_LINE:
			reg = AC97REG_LINE_IN;
			break;
		case SOUND_MIXER_LINE1:
			reg = AC97REG_AUX;
			break;
		case SOUND_MIXER_VIDEO:
			reg = AC97REG_VIDEO;
			break;
		case SOUND_MIXER_CD:
		default:
			reg = AC97REG_CD;
			break;
		}
		if (left < 7 && right < 7) {
			WriteCodec(dev->core, reg, 0x8000);
			break;
		}
		if (left < 7)
			left = 7;
		if (right < 7)
			right = 7;
		WriteCodec(dev->core, reg, (((100 - left) / 3) << 8) | ((100 - right) / 3));
		break;
	case SOUND_MIXER_TREBLE:
	case SOUND_MIXER_BASS:
		/* fixme: check codec caps, use EQ block */
		break;
	case SOUND_MIXER_PHONEOUT:
	case SOUND_MIXER_IGAIN:
		/* fixme: check codec caps */
		break;
	case SOUND_MIXER_MIC:
		if (left < 11) {
			WriteCodec(dev->core, AC97REG_MIC, 0x8000);
			break;
		}
		boost = 0;
		if (left >= 27) {
			left -= 27;
			boost = 0x40;
		}
		if (left < 11)
			left = 11;
		WriteCodec(dev->core, AC97REG_MIC, ((73 - left) / 2) | boost);
		break;
	case SOUND_MIXER_SPEAKER:
		WriteCodec(dev->core, AC97REG_PC_BEEP, (left < 10) ? 0x8000 : ((100 - left) / 6) << 1);
		break;
	case SOUND_MIXER_OGAIN:
		WriteCodec(dev->core, AC97REG_MASTER_MONO, (left < 7) ? 0x8000 : (100 - left) / 3);
		break;
	case SOUND_MIXER_PHONEIN:
		WriteCodec(dev->core, AC97REG_PHONE, (left < 7) ? 0x8000 : (100 - left) / 3);
		break;
	case SOUND_MIXER_RECLEV:
		if (left < 10 || right < 10) {
			WriteCodec(dev->core, AC97REG_REC_GAIN, 0x8000);
			WriteMixer(dev->core, ASP_MIX_INPUT_WAVE_AT_WAVEIN, 0, 0);
			break;
		}
		if (left < 10)
			left = 10;
		if (right < 10)
			right = 10;
		WriteCodec(dev->core, AC97REG_REC_GAIN, (((left - 10) / 6) << 8) | ((right - 10) / 6));
		WriteMixer(dev->core, ASP_MIX_INPUT_WAVE_AT_WAVEIN, (left * 655), (right * 655));
		break;
	default:
		//PWARN("write mixer: unknown mixer chan %d\n", (int)chan);
		break;
	}
}

static int au_mixer_open(struct inode *inode, struct file *file)
{
	au_device_t *dev;
	int minor_num = MINOR(inode->i_rdev);

	//PDEBUG("au_mixer_open, minor = %d\n", minor_num);

	dev = find_device(minor_num);
	if (!dev) {
		PWARN("could not find device for minor = %d\n", minor_num);
		return -ENODEV;
	}
	file->private_data = (void *)dev;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	MOD_INC_USE_COUNT;
#endif
	return SUCCESS;
}

static int au_mixer_release(struct inode *inode, struct file *file)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	MOD_DEC_USE_COUNT;
#endif
	return SUCCESS;
}

static void init_mixer(au_device_t *dev)
{
	int i;
	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++) {
		dev->mix_vol[i] = 0x5050;
		write_mixer(dev, i, dev->mix_vol[i]);
	}

	dev->mix_recsrc = SOUND_MASK_MIC;
	set_rec_src(dev, dev->mix_recsrc);		// default to mic source
}

static void set_rec_src(au_device_t *dev, unsigned int recsrc)
{
	short val;

	switch (recsrc) {
	case SOUND_MASK_CD:
		val = 0x1;
		break;
	case SOUND_MASK_VIDEO:
		val = 0x2;
		break;
	case SOUND_MASK_LINE1:	//aux?
		val = 0x3;
		break;
	case SOUND_MASK_LINE:
		val = 0x4;
		break;
	case SOUND_MASK_VOLUME:
		val = 0x5;
		break;
	case SOUND_MASK_PHONEOUT:	//mono mix?
		val = 0x6;
		break;
	case SOUND_MASK_PHONEIN:
		val = 0x7;
		break;
	case SOUND_MASK_MIC:
	default:
		val = 0;
		break;
	}
	WriteCodec(dev->core, AC97REG_REC_SELECT, (val << 8 | val));
}
