/*
 * au_sndstat.c
 *
 * Aureal Vortex AU88xx audio driver module
 * /dev/sndstat Functions
 * 
 * Copyright (c) 1999 Aureal Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Some of this code is based on the es1371 driver, which is
 * Copyright (C) 1998  Thomas Sailer (sailer@ife.ee.ethz.ch)
 * 
 * Revision History:
 *
 * 19-08-2000 Nikos Kalogridis
 *      Split vortex.c into multiple files for easier maintenance
 * 
 * 24-08-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Reformatted Code
 *
 */

#define __NO_VERSION__
#include "au_vortex.h"
#include <linux/version.h>

/* file operations */

static int au_sndstat_open(struct inode *inode, struct file *file);
static int au_sndstat_release(struct inode *inode, struct file *file);
static ssize_t au_sndstat_read(struct file *file, char *buffer, size_t count, loff_t *ppos);
static int sndstat_file_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos);
static int sound_proc_get_info(char *buffer, char **start, off_t offset, int length, int inout);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
static struct file_operations sndstat_fops =
{
	owner:		THIS_MODULE,
	read:		au_sndstat_read,
	open:		au_sndstat_open,
	release:	au_sndstat_release,
};
#else
static struct file_operations sndstat_fops = {
        NULL,                   /* llseek */
        au_sndstat_read,
        NULL,                   /* write */
        NULL,                   /* readdir */
        NULL,                   /* poll */  
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        au_sndstat_open,
        NULL,                   /* flush */
        au_sndstat_release,
        NULL,                   /* fsync */ 
        NULL,                   /* fasync */
        NULL,                   /* check media change */
        NULL,                   /* revalidate */
        NULL,                   /* lock */  
};
#endif

/********************************/
/* sndstat  support		*/
/********************************/

/*
 * Some of this sndstat code is from soundcard.c, which is
 * Copyright 1993-1997 Hannu Savolainen
 */

/* local variables to store if there is a device registered */
/* and also which */
static int dev_sndstat = -1;
static au_device_t *registered_dev = NULL;

int au_sndstat_register(au_device_t *dev)
{
	int status;
	
	/*
	 * There is only one sndstat device. Register it if we can.
	 */
	PDEBUG("au_sndstat_register\n");
	if (dev_sndstat < 0) {
		status = dev_sndstat = register_sound_special(&sndstat_fops, 6);
		if(status < 0)
			return -1;
		registered_dev = dev;
	}
	return 0;
}	

int au_sndstat_unregister(au_device_t *dev)
{
	PDEBUG("au_sndstat_unregister\n");
	if ((dev_sndstat >= 0) && ((dev = registered_dev) || (dev == NULL))) {
		dev_sndstat = -1;
		registered_dev = NULL;
		unregister_sound_special(6);
	}
	return 0;
}

static int au_sndstat_open(struct inode *inode, struct file *file)
{
	PDEBUG("au_sndstat_open\n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	MOD_INC_USE_COUNT;
#endif
	return SUCCESS;
}

static ssize_t au_sndstat_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	PDEBUG("au_sndstat_read, count=%d\n", count);
	return sndstat_file_read(file, buffer, count, &file->f_pos);
}

/* 4K page size but our output routines use some slack for overruns */
#define PROC_BLOCK_SIZE (3*1024)

/*
 * basically copied from fs/proc/generic.c:proc_file_read 
 * should be removed sometime in the future together with /dev/sndstat
 * (a symlink /dev/sndstat -> /proc/sound will do as well)
 */
static ssize_t sndstat_file_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos)
{
	char *page;
	ssize_t	retval=0;
	int eof=0;
	ssize_t	n, count;
	char *start;

	if (!(page = (char*) __get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	while ((nbytes > 0) && !eof) {
		count = MIN(PROC_BLOCK_SIZE, nbytes);
		start = NULL;
		n = sound_proc_get_info(page, &start, *ppos, count, 0);
		if (n < count)
			eof = 1;
                        
		if (!start) {
			/* For proc files that are less than 4k */
			start = page + *ppos;
			n -= *ppos;
			if (n <= 0)
				break;
			if (n > count)
				n = count;
		}
		if (n == 0)
			break;  /* End of file */
		if (n < 0) {
			if (retval == 0)
				retval = n;
			break;
		}
                
		n -= copy_to_user(buf, start, n);       /* BUG ??? */
		if (n == 0) {
			if (retval == 0)
				retval = -EFAULT;
			break;
		}
                
		*ppos += n;     /* Move down the file */
		nbytes -= n;
		buf += n;
		retval += n;
	}
	free_page((unsigned long) page);
	return retval;
}

int sound_proc_get_info(char *buffer, char **start, off_t offset, int length, int inout)
{
	int len, count;
	off_t begin = 0;
	au_device_t *dev;

#ifdef MODULE
#define MODULEPROCSTRING "Driver loaded as a module"
#else
#define MODULEPROCSTRING "Driver compiled into kernel"
#endif

	if (!dev_head)
		return 0;

	len = sprintf(buffer, "OSS/Free: 0x%x\n"
		      "Load type: " MODULEPROCSTRING "\n"
		      "Kernel: %s %s %s %s %s\n"
		      "Config options: %x\n\nInstalled drivers: \n", SOUND_VERSION,
		      system_utsname.sysname, system_utsname.nodename, system_utsname.release,
		      system_utsname.version, system_utsname.machine, 0);
	len += sprintf(buffer + len, "Type 0: %s\n", "Aureal Vortex");

	len += sprintf(buffer + len, "\nCard config: \n");
	dev = dev_head;
	while (dev) {
		len += sprintf(buffer + len, "(");
		len += sprintf(buffer + len, "%s", sound_proc_get_chip(dev->dev_id));
		len += sprintf(buffer + len, " at 0x%lx", RSRCADDRESS(dev->pcidev, 0));
		len += sprintf(buffer + len, " irq %d", dev->pcidev->irq);
		len += sprintf(buffer + len, ")");
		len += sprintf(buffer + len, "\n");
		dev = dev->next;
	}

	len += sprintf(buffer + len, "\nAudio devices:\n");
	dev = dev_head;
	count = 0;
	while (dev) {
		len += sprintf(buffer + len, "%d: %s%s\n", count++, sound_proc_get_chip(dev->dev_id), " (DUPLEX)");
		dev = dev->next;
	}

	len += sprintf(buffer + len, "\nSynth devices: \nNot supported by current driver\n");

	len += sprintf(buffer + len, "\nMidi devices: \n");
	dev = dev_head;
	count = 0;
	while (dev) {
		len += sprintf(buffer + len, "%d: %s\n", count++, sound_proc_get_chip(dev->dev_id));
		dev = dev->next;
	}

	len += sprintf(buffer + len, "\nTimers:\n");
	len += sprintf(buffer + len, "0: %s\n", "System Clock");

	len += sprintf(buffer + len, "\nMixers:\n");
	dev = dev_head;
	count = 0;
	while (dev) {
		len += sprintf(buffer + len, "%d: %s\n", count++, sound_proc_get_chip(dev->dev_id));
		dev = dev->next;
	}

	*start = buffer + (offset - begin);
	len -= (offset - begin);
	if (len > length) 
		len = length;
	return len;
}

static int au_sndstat_release(struct inode *inode, struct file *file)
{
	PDEBUG("au_sndstat_release\n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	MOD_DEC_USE_COUNT;
#endif
	return SUCCESS;
}
