/*
 * au_vortex.c
 *
 * Aureal Vortex AU88xx audio driver module.
 * Kernel/PCI Interface
 *
 * Copyright (c) 1999 Aureal Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Contact: http://linux.aureal.com
 *
 * Some of this code is based on the es1371 driver, which is
 * Copyright (C) 1998  Thomas Sailer (sailer@ife.ee.ethz.ch)
 *
 * This file must be linked with the appropriate Vortex core object file
 * (asp30.o, asp10.o, or asp20.o).
 *
 * Devices:
 * /dev/dsp	full duplex, multiple streams
 * /dev/mixer
 * /dev/midi
 * /dev/audio
 * /dev/sndstat
 *
 * The joystick interface is enabled at I/O port 0x201.
 * Joystick support can be provided by the standard joystick driver.
 *
 * Revision History:
 *
 * 11-01-1999 Michael Minnick
 *	Initial release
 *
 * 02-10-2000 Michael Minnick
 *	Fill entire dma buf before sending
 *
 * 19-03-2000 Eduardo G. Andrade
 *	Stability changes
 *
 * 15-07-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Now enjoys compiling (and working!) on all kernel versions
 *
 * 31-07-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Fixed module use counts for 2.4.x kernels
 *
 * 10-08-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Now compiles on 2.4.0test6 - Previous test versions depreciated
 *
 * 12-08-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Fixed lockups on test6, misc code cleanup
 *
 * 18-08-2000 Nikos Kalogridis
 *	Converted to new PCI device driver API
 *
 * 19-08-2000 Nikos Kalogridis
 *      Split vortex.c into multiple files for easier maintenance
 *
 * 23-08-2000 Derek Foreman
 *	Merged changes from mminick's 1.0.6 drivers (spdif)
 *
 * 24-08-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Reformatted Code
 *
 * 25-08-2000 Philipp Hahn
 *	Added MODULE_DESCRIPTION/MODULE_AUTHOR
 *
 */

#include "au_vortex.h"

/* local functions */
/* The following functions are not used in the new pci style driver so there */
/* is no need to have them.             */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
static int device_scan(void);
static void remove_devices(void);
static int scan_id(int dev_id);
#endif /* Old PCI driver style functions */

static void au_isr(int irq, void *dev_id, struct pt_regs *regs);
static int add_device(struct pci_dev *pcidev);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
/********************************/
/* PCI support                  */
/********************************/

static struct pci_device_id aureal_pci_tbl[] __devinitdata = {
	{VENDOR_ID, DEV_AU8810, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DEV_AU8810 },
	{VENDOR_ID, DEV_AU8820, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DEV_AU8820 },
	{VENDOR_ID, DEV_AU8830, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DEV_AU8830 },
	{0,}
};
MODULE_DEVICE_TABLE (pci, aureal_pci_tbl);

static int __devinit aureal_init_one (struct pci_dev *pdev,
				       const struct pci_device_id *ent)
{
	int temp;

	PDEBUG("init one called\n");
#ifdef AU8810
	if (pdev->device != DEV_AU8810) {
		PWARN("Driver compiled with AU8810 option.\n");
#endif
#ifdef AU8820
	if (pdev->device != DEV_AU8820) {
		PWARN("Driver compiled with AU8820 option.\n");
#endif
#ifdef AU8830
	if (pdev->device != DEV_AU8830) {
		PWARN("Driver compiled with AU8830 option.\n");
#endif
		PWARN("Found Chipset id: 0x%04x\n", pdev->device);
		PWARN("It appears that you compiled the driver with the wrong chipset option.\n");
		PWARN("Please change the Makefile of the driver and compile again to use this\n");
		PWARN("chipset.\n");
		return -EIO;
	}
	PDEBUG("card resources before pci_enable_device call\n");
	PDEBUG("id=0x%04x\n", pdev->device);
	PDEBUG("bar0=0x%08lx\n", RSRCADDRESS(pdev, 0));
	PDEBUG("irq=%d\n", pdev->irq);

	/* Check if card is already initialized by BIOS and if not warn the user */
	if((RSRCADDRESS(pdev, 0) == 0) ||
	  ((RSRCADDRESS(pdev, 0) & PCI_BASE_ADDRESS_SPACE) != PCI_BASE_ADDRESS_SPACE_MEMORY) ||
	   (pdev->irq == 0)) {
		PWARN("found uninitialized vortex card!\n");
		PWARN("This driver is not expected to work at this time with an uninitialized\n");
		PWARN("Vortex card. Please look in the BIOS Settings for a PNP OS option and\n");
		PWARN("set it to NO (or Disabled) to make your card work at this time.\n");
		PWARN("If by any chance the card IS working though please send an email to:\n");
		PWARN("nikos_kalogridis@hotmail.com with subject VORTEX LINUX.\n");
	}
	
	if (SUCCESS != (temp = pci_enable_device(pdev))) {
		PWARN("Failed to enable device!\n");
		return temp;
	}
   	pci_set_master(pdev);
   	PDEBUG("card resources after pci_enable_device call\n");
	PWARN("Found vortex PCI device:\n");
	PWARN("id=0x%04x\n", pdev->device);
	PWARN("bar0=0x%08lx\n", RSRCADDRESS(pdev, 0));
	PWARN("irq=%d\n", pdev->irq);

	if(SUCCESS != (temp = add_device(pdev)))
		PWARN("failed to add device\n");

	return temp;
}

static void __devexit aureal_remove_one (struct pci_dev *pdev)
{
	au_device_t *dev = NULL;

	PDEBUG("remove one called\n");
	/* Locate the au_dev that holds the pci_dev to be removed and */
	/* remove it from the list */
	if (!(dev = find_device_by_pcidev(pdev))) {
		PWARN("Device requested to be removed was not found in the list!\n");
		return;
	}
	remove_device_from_list(dev);
	/* Now release the resources for that device */
	coreDisableInts(dev->core);
	free_irq(dev->pcidev->irq, dev);
	release_mem_region(RSRCADDRESS(dev->pcidev,0),0x40000);
	release_region(RSRCADDRESS(dev->pcidev,1),8);
	release_region(RSRCADDRESS(dev->pcidev,2),8);
	/* HwFreeSynth() */
	deallocCore(dev->core);
	au_dsp_unregister(dev);
	au_mixer_unregister(dev);
	au_midi_unregister(dev);
	au_audio_unregister(dev);
	au_sndstat_unregister(dev);
	kfree(dev);
}

static void aureal_resume (struct pci_dev *pdev)
{
	PDEBUG("resume called\n");
}

static void aureal_suspend (struct pci_dev *pdev)
{
	PDEBUG("suspend called\n");
}

static struct pci_driver aureal_pci_driver = {
	name:		DRIVER_NAME,
	id_table:	aureal_pci_tbl,
	probe:		aureal_init_one,
	remove:		aureal_remove_one,
	suspend:	NULL, /* for now instead of aureal_suspend,*/
	resume:		NULL, /* for now instead of aureal_resume, */
};

/********************************/
/* Module support		*/
/********************************/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
MODULE_DESCRIPTION("Aureal Vortex AU88xx PCI Audio Driver");
MODULE_AUTHOR("Eduardo G. Andrade, Warren Chartier, Derek Foreman, Nikos Kalogridis, Michael Minnick");
#endif

static int __init aureal_init_module (void)
{
	EXPORT_NO_SYMBOLS;
	PWARN("Loading...\n");
	return pci_module_init (&aureal_pci_driver);
}

static void __exit aureal_cleanup_module (void)
{
	pci_unregister_driver (&aureal_pci_driver);
	dev_head = dev_tail = (au_device_t*)NULL;
	PWARN("Unloaded.\n");
}

module_init(aureal_init_module);
module_exit(aureal_cleanup_module);

#else /* Old PCI driver style */

int init_module()
{
	EXPORT_NO_SYMBOLS;
	PWARN("Loading...\n");
	/* scan pci bus and create device structures */
	return device_scan();
}

void cleanup_module()
{
	remove_devices();
	PWARN("Unloaded.\n");
}

/********************************/
/* Device support		*/
/********************************/

static int device_scan()
{
	int device_count = 0;

	if(!pci_present())
		return -ENODEV;
	/*
	 * This driver binary supports multiple instances of vortex
	 * cards of one particular chip type.
	 */
#ifdef AU8810
	device_count += scan_id(DEV_AU8810);
#endif
#ifdef AU8820
	device_count += scan_id(DEV_AU8820);
#endif
#ifdef AU8830
	device_count += scan_id(DEV_AU8830);
#endif
	return device_count ? SUCCESS : -ENODEV;
}

static int scan_id(int dev_id)
{
	struct 	pci_dev*	pcidev 		= NULL;
	int 			device_count 	= 0;

	PWARN("Scanning for device VENDOR=0x%x, ID=0x%x\n", VENDOR_ID, dev_id);

	while((pcidev = pci_find_device(VENDOR_ID, dev_id, pcidev)) != 0) {
		PWARN("Found vortex PCI device:\n");
		PWARN("id=%d\n", dev_id);
		PWARN("bar0=0x%08lx\n", RSRCADDRESS(pcidev, 0));
		PWARN("irq=%d\n", pcidev->irq);

		if((RSRCADDRESS(pcidev, 0) == 0) ||
		  ((RSRCADDRESS(pcidev, 0) & PCI_BASE_ADDRESS_SPACE) != PCI_BASE_ADDRESS_SPACE_MEMORY) ||
		   (pcidev->irq == 0)) {
			PWARN("found uninitialized vortex card\n");
			continue;
		}
		if(add_device(pcidev) == 0)
			device_count++;
	}
	return device_count;
}

static void remove_devices()
{
	au_device_t *next_dev = dev_head;
	au_device_t *dev;

	while(next_dev) {
		dev = next_dev;
		next_dev = dev->next;
		coreDisableInts(dev->core);
		free_irq(dev->pcidev->irq, dev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
		release_mem_region(RSRCADDRESS(dev->pcidev,0),0x40000);
#endif
		release_region(RSRCADDRESS(dev->pcidev,1),8);
		release_region(RSRCADDRESS(dev->pcidev,2),8);
		/* HwFreeSynth() */
		deallocCore(dev->core);
		au_dsp_unregister(dev);
		au_mixer_unregister(dev);
		au_midi_unregister(dev);
		au_audio_unregister(dev);
		kfree(dev);
	}
	dev_head = dev_tail = (au_device_t*)NULL;
	au_sndstat_unregister(NULL);
}

#endif /* Old PCI driver style functions */

static void release_resources(au_device_t *dev, int level)
{
	switch (level) {
		case 6:
			au_audio_unregister(dev);
		case 5:
			au_midi_unregister(dev);
		case 4:
			au_mixer_unregister(dev);
		case 3:
			au_dsp_unregister(dev);
		case 2:
			coreDisableInts(dev->core);
			free_irq(dev->pcidev->irq, dev);
		case 1:
			deallocCore(dev->core);
		case 0:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
			release_mem_region(RSRCADDRESS(dev->pcidev,0),0x40000);
#endif
			release_region(RSRCADDRESS(dev->pcidev,1),8);
			release_region(RSRCADDRESS(dev->pcidev,2),8);
			kfree(dev);
	}
}

static int add_device(struct pci_dev *pcidev)
{
	au_device_t *dev;
	int status;

	dev = (au_device_t *)kmalloc(sizeof(au_device_t), GFP_KERNEL);
	if(!dev)
		return -ENOMEM;
	memset(dev, 0, sizeof(au_device_t));
	spin_lock_init(&dev->irq_lock);
	dev->dev_id = pcidev->device;
	dev->pcidev = pcidev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
	init_waitqueue_head(&dev->midi.iwait);
	init_waitqueue_head(&dev->midi.owait);
#else
	init_waitqueue(&dev->midi.iwait);
	init_waitqueue(&dev->midi.owait);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
	if(check_mem_region(RSRCADDRESS(dev->pcidev,0),0x40000))
		PWARN("mem %#lx-%#lx in use\n", RSRCADDRESS(dev->pcidev,0),RSRCADDRESS(dev->pcidev,0)+0x3FFFF);
	else
#endif
		if(check_region(RSRCADDRESS(dev->pcidev,1),8))
	    	PWARN("io ports %#lx-%#lx in use\n", RSRCADDRESS(dev->pcidev,1),RSRCADDRESS(dev->pcidev,1)+8);
	  	else
	    	if(check_region(RSRCADDRESS(dev->pcidev,2),8))
	      		PWARN("io ports %#lx-%#lx in use\n", RSRCADDRESS(dev->pcidev,2),RSRCADDRESS(dev->pcidev,2)+7);
	    	else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
				/* QUESTION: Shouldn't we check these funcs if they succeed or fail? */
	      		request_mem_region(RSRCADDRESS(dev->pcidev,0),0x40000,DRIVER_NAME);
#endif
	      		request_region(RSRCADDRESS(dev->pcidev,1),8,DRIVER_NAME);
	      		request_region(RSRCADDRESS(dev->pcidev,2),8,DRIVER_NAME);
	    	}
	
  	/* allocate the vortex core */
	if ((dev->core = alloc_core(pcidev, dev)) == NULL) {
		PWARN("Could not allocate core!\n");
		release_resources(dev, 0);
		return -ENOMEM;
	}

	if (request_irq(dev->pcidev->irq, au_isr, SA_SHIRQ, "au88xx", dev) != 0) {
		PWARN("Cannot alloc irq!\n");
		release_resources(dev, 1);
		return -ENOMEM;
	}


	/* fixme alloc synth - HwAllocSynth */

	/* fixme - pass port as driver arg for each card, allow disable */
	EnableHardCodedJoystickPort(dev->core, TRUE, 0x201);

	coreEnableInts(dev->core);

	/* register devices with soundcore */
	if (SUCCESS != (status = au_dsp_register(dev))) {
		PWARN("Could not register audio device!\n");
		release_resources(dev, 2);
		return status;
	}
	if (SUCCESS != (status = au_mixer_register(dev))) {
		PWARN("Could not register mixer device!\n");
		release_resources(dev, 3);
		return status;
	}
	if (SUCCESS != (status = au_midi_register(dev))) {
		PWARN("Could not register midi device!\n");
		release_resources(dev, 4);
		kfree(dev);
		return status;
	}
	/* this is /dev/audio */
	if (SUCCESS != (status = au_audio_register(dev))) {
		PWARN("Could not register sun audio device!\n");
		release_resources(dev, 5);
		return status;
	}
	if (SUCCESS != (status = au_sndstat_register(dev)))
		PWARN("Could not register sndstat device!\n");
	
	/* add new device to our device list */
	add_device_to_list(dev);

	PWARN("Add device, audio=%d, mixer=%d, midi=%d\n", dev->dev_audio, dev->dev_mixer, dev->dev_midi);

	return SUCCESS;
}

static void au_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	au_device_t *dev = (au_device_t *)dev_id;
	unsigned long irq_status;
	//PDEBUG("AU_ISR - dev = 0x%x\n", (unsigned int)dev);
	/* Calls into the core outside this ISR are protected by this spin lock. */
	/* It also protects the dev data structure within this module. */
	spin_lock(&dev->irq_lock);

	/* the core ISR acknowledges the irq and returns the mask */
	irq_status = coreISR(dev->core);
	//PDEBUG("irq_s = 0x%x\n", (unsigned int)irq_status);
	/* shared interrupt handler - return if not ours */
	if(!irq_status) {
		spin_unlock(&dev->irq_lock);
		return;
	}
	if(irq_status & 0x2000)
		au_handle_midi(dev);
	if(irq_status & 0x1000)
		au_handle_timer(dev);
	spin_unlock(&dev->irq_lock);
}
