/*
 * au_vortex.h
 *
 * Aureal Vortex AU88xx audio driver module
 * Public Declarations and Structures
 *
 * Copyright (c) 1999 Aureal Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Some of this code is based on the es1371 driver, which is
 * Copyright (C) 1998  Thomas Sailer (sailer@ife.ee.ethz.ch)
 * 
 * Revision History:
 *
 * 19-08-2000 Nikos Kalogridis
 *      Split vortex.c into multiple files for easier maintenance
 *
 * 20-08-2000 Nikos Kalogridis
 *      Exposed au_mixer_ioctl function to be used by au_audio.c
 *
 * 24-08-2000 Warren Chartier <icebalm@bigfoot.com>
 *	Reformatted Code
 * 
 */

#include <linux/config.h>
#if defined(CONFIG_SMP) && !defined(__SMP__)
#define __SMP__
#endif

#include <linux/version.h>
 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 1, 0)
#error "This driver does not support pre-2.1 kernels"
#endif

#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/sound.h>
#include <linux/malloc.h>
#include <linux/soundcard.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/utsname.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
#include <linux/spinlock.h>
#else
#include <asm/spinlock.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
#include <linux/wrapper.h>
#include <linux/init.h>
#endif

#define DRIVER_NAME 	"au88xx"

#define VENDOR_ID	0x12eb
/* pci device ids */
#define DEV_AU8810	3
#define DEV_AU8830	2
#define DEV_AU8820	1

#define PWARN(fmt, args...) printk(KERN_WARNING DRIVER_NAME ": " fmt, ## args)
#ifdef DEBUG_AU
#define PDEBUG(fmt, args...) printk(KERN_DEBUG DRIVER_NAME ": " fmt, ## args)
#else
#define PDEBUG(fmt, args...)
#endif

#ifdef __VerboseMode
#define PSHOW(fmt, args...) printk(KERN_WARNING DRIVER_NAME "->" fmt, ## args)
#else
#define PSHOW(fmt, args...)
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define SUCCESS 0

#ifndef MIN
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
#	define RSRCADDRESS(dev,num) ((dev)->resource[(num)].start)
#else
#	define RSRCADDRESS(dev,num) ((dev)->base_address[(num)])
#endif

/*
 * Memory transfer buffer.
 * Each stream contains one or more of these buffers,
 * which are used to transfer application buffers
 * to the vortex hardware.
 */
typedef struct mem_buf {
	unsigned long	count;
	unsigned long	xfer;
	void		*page;
	unsigned	in_use: 1;
} mem_buf_t;

#define DEF_FRAG_SIZE	PAGE_SIZE
#define MAX_FRAGS	4
#define DEF_NR_FRAGS	MAX_FRAGS

/*
 * Wave stream.
 * Either a playback or a record stream.
 */
typedef struct wave_stream {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
        wait_queue_head_t       io_wait;
#else
        struct wait_queue       *io_wait;
#endif
	void			*wave;		/* core wave object pointer */
	void			*dmabuf;
	mem_buf_t		buf[MAX_FRAGS];
	unsigned long		total_played;
	unsigned long		total_sent;
	unsigned long		frag_size;
	unsigned long		num_frags;
	int			buf_order;
	int			fill_index;
	int			drain_index;
	int			data_format;
	int			sample_rate;
	int			channel_count;
	unsigned		active: 1;
	unsigned		pre_roll: 1;
	unsigned		mapped: 1;
	unsigned		mapped_setup: 1;
} wave_stream_t;

/*
 * Wave device.
 * Contains a playback stream and a record stream (full duplex).
 */
typedef struct wave_device {
	struct wave_device	*next;
	struct wave_device	*prev;
	struct file		*fp;
	wave_stream_t		play_stream;
	wave_stream_t		record_stream;
} wave_device_t;

/* MIDI buffer sizes */
#define MIDIINBUF  256
#define MIDIOUTBUF 256

/* vortex gameport midi status register */
#define MPU401_CMDOK	0x40        /* ok to write if 0 */
#define MPU401_MIDIVAL	0x80        /* data avail if 0 */

/*
 * Midi device.
 * Contains an input buffer and an output buffer.
 */
typedef struct midi_device {
	unsigned		ird, iwr, icnt;
	unsigned		ord, owr, ocnt;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 30)
        wait_queue_head_t       iwait;
        wait_queue_head_t       owait;
#else
        struct wait_queue       *iwait;
        struct wait_queue       *owait;
#endif
	unsigned char		ibuf[MIDIINBUF];
	unsigned char		obuf[MIDIOUTBUF];
} midi_device_t;

/*
 * Per-device state.
 * Each device represents a vortex chip, typically on a pci card.
 * Each device has its own vortex core instance.
 */
typedef struct au_device {
	struct au_device	*next;
	struct pci_dev		*pcidev;
	void			*core;		/* core object pointer */
	int			dev_id;		/* vortex chip id */
	int			dev_audio;	/* audio (dsp) minor num from soundcore */
	int			dev_mixer;
	int			dev_midi;
	int			dev_sun;
	int			mix_modcnt;
	unsigned int		mix_recsrc;
	short			mix_vol[SOUND_MIXER_NRDEVICES];
	spinlock_t		irq_lock;
	wave_device_t		*wave_head;
	wave_device_t		*wave_tail;
	midi_device_t		midi;
	unsigned		record_in_use: 1;
	unsigned		midiin_in_use: 1;
	unsigned		midiout_in_use: 1;
} au_device_t;

/*
 * Each device has multiple "functions" - audio, mixer, midi, synth.
 * Functions are assigned a minor number by the soundcore module.
 * The lower nibble is the function type.
 * For example /dev/dsp minor number is 3 and would be
 * assigned to the audio (wave) function on the first vortex card.
 * Minor numbers are "chained" using the high nibble, so the
 * second device would get /dev/dsp1, whose minor number is 19.
 * That's 0x13, low nibble = 3 (dsp), high nibble = 1 (chain).
 */
#define DEVICE_TYPE(m)	(m & 0x0f)
#define DEV_MIXER		0
#define DEV_SEQUENCER		1
#define DEV_MIDI		2
#define DEV_AUDIO		3
#define DEV_SUNDSP		4

/* Audio interface functions */

void au_handle_timer(au_device_t *dev);
int au_dsp_register(au_device_t *dev);
int au_dsp_unregister(au_device_t *dev);
int au_audio_register(au_device_t *dev);
int au_audio_unregister(au_device_t *dev);

/* Midi interface functions */

void au_handle_midi(au_device_t *dev);
int au_midi_register(au_device_t *dev);
int au_midi_unregister(au_device_t *dev);

/* Mixer interface functions */

int au_mixer_register(au_device_t *dev);
int au_mixer_unregister(au_device_t *dev);
int au_mixer_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

/* SndStat interface functions */

int au_sndstat_register(au_device_t *dev);
int au_sndstat_unregister(au_device_t *dev);

/* These two variables should be hidden in the au_utils.c but we leave */
/* them here because the au_sndstat module access them directly. FIXME */
au_device_t *dev_head;	/* list of device found on pci bus */
au_device_t *dev_tail;

/* Utility functions */
au_device_t *find_device(int minor_num);
au_device_t *find_device_by_pcidev(struct pci_dev *pdev);
int add_device_to_list(au_device_t *dev);
int remove_device_from_list(au_device_t *dev);

loff_t au_llseek(struct file *file, loff_t offset, int origin);

char *sound_proc_get_chip(int dev_id);

/********************************/
/* Vortex Core interface	*/
/********************************/

/* We're not including core header files, so we must count on
   these interfaces not changing. */

typedef enum {
	ASPFMTLINEAR16 = 0,
	ASPFMTLINEAR8,
	ASPFMTULAW, 
	ASPFMTALAW, 
	ASPFMTSPORT,
	ASPFMTSPDIF,
} ASPENCODING;

typedef enum {
	ASPDIRPLAY,
	ASPDIRRECORD
} ASPDIRECTION;

typedef struct _ASPWAVEFORMAT {
	unsigned short	wBitWidth;
	unsigned short	wChannels;
	unsigned long	dwSampleRate;
	ASPENCODING	eEncoding;
} ASPWAVEFORMAT, *PASPWAVEFORMAT;

typedef struct _ASPVOLUME {
	long	lVolume;
	long	lPanPosition;
} ASPVOLUME, *PASPVOLUME;

#define	MAX_MEM_REGISTERS	9
#define	MAX_IO_PORTS		20
#define	MAX_IRQS		7
#define	MAX_DMA_CHANNELS	7

typedef struct Config_Buff_s {
	unsigned short	wNumMemWindows;
	unsigned long	dMemBase[MAX_MEM_REGISTERS];
	unsigned long	dMemLength[MAX_MEM_REGISTERS];
	unsigned short	wMemAttrib[MAX_MEM_REGISTERS];
	unsigned short	wNumIOPorts;
	unsigned short	wIOPortBase[MAX_IO_PORTS];
	unsigned short	wIOPortLength[MAX_IO_PORTS];
	unsigned short	wNumIRQs;
	unsigned char	bIRQRegisters[MAX_IRQS];
	unsigned char	bIRQAttrib[MAX_IRQS];
	unsigned short	wNumDMAs;
	unsigned char	bDMALst[MAX_DMA_CHANNELS];
	unsigned short	wDMAAttrib[MAX_DMA_CHANNELS];
	unsigned char	bReserved1[3];
	unsigned long	BusNumber;
} CMCONFIG;

/* AC97 codec registers */
#define AC97REG_MASTER			0x02
#define AC97REG_MASTER_MONO		0x06
#define AC97REG_PC_BEEP			0x0a
#define AC97REG_PHONE			0x0c
#define AC97REG_MIC			0x0e
#define AC97REG_LINE_IN			0x10
#define AC97REG_CD			0x12
#define AC97REG_VIDEO			0x14
#define AC97REG_PCM			0x18
#define AC97REG_AUX			0x16
#define AC97REG_REC_GAIN		0x1c
#define AC97REG_REC_SELECT		0x1a

/* mixer channels */
#define ASP_MIX_INPUT_WAVE_AT_PCM	12
#define ASP_MIX_INPUT_WAVE_AT_WAVEIN	14

typedef void  ClassAsp4Core;
typedef void  ClassWave;

extern void *allocCore(CMCONFIG *pConfig);
extern void deallocCore(ClassAsp4Core *pCore);
extern unsigned long coreISR(ClassAsp4Core *pCore);
extern void coreEnableInts(ClassAsp4Core *pCore);
extern void coreDisableInts(ClassAsp4Core *pCore);
extern void EnableHardCodedJoystickPort(ClassAsp4Core *pCore, int bEnable, unsigned long dwPort);
extern int CreateWaveBuffer(ClassAsp4Core *pCore, PASPWAVEFORMAT pWaveFormat, ASPDIRECTION eDirection, ClassWave **ppWaveObj);
extern void StopWave(ClassWave *pWave);
extern void ReleaseWave(ClassWave *pWave);
extern void StartWave(ClassWave *pWave);
extern unsigned long GetWaveBytesPlayed(ClassWave *pWave);
extern int AddWaveBuffer(ClassWave *pWave, void *buf, unsigned long size);
extern void SetWaveFormat(ClassWave *pWave, PASPWAVEFORMAT pWaveFormat);
extern void WriteCodec(ClassAsp4Core *pCore, short reg, short val);
extern void ReadCodec(ClassAsp4Core *pCore, short reg, unsigned short *val);
extern void WriteMixer(ClassAsp4Core *pCore, unsigned long id, unsigned long left, unsigned long right);
extern void EnableHardCodedMPU401Port(ClassAsp4Core *pCore, int bEnable, unsigned long dwPort);
extern void EnableMPU401Interrupt(ClassAsp4Core *pCore, int bEnable);
extern void Mpu401UartWriteData(ClassAsp4Core *pCore, unsigned char bData);
extern unsigned char Mpu401UartReadData(ClassAsp4Core *pCore);
extern unsigned char Mpu401UartReadStatus(ClassAsp4Core *pCore);
extern void Mpu401UartWriteCommand(ClassAsp4Core *pCore, unsigned char bData);
extern int Mpu401UartInit(ClassAsp4Core *pCore, unsigned char bClockDivider, unsigned char bMidiMode, int bMpuMode);

void *alloc_core(struct pci_dev *pcidev, au_device_t *dev);
